## **Git Workflow Guidance**


### Overview

1. [Tracking Changes](#1-tracking-changes)
2. [Writing Good Commit Messages](#2-writing-good-commit-messages)
3. [Handling Merge Conflicts](#3-handling-merge-conflicts)
4. [Using a `.gitignore` File](#4-using-a-gitignore-file)
5. [Undoing Mistakes](#5-undoing-mistakes)
6. [Saving Work Temporarily with `git stash`](#6-saving-work-temporarily-with-git-stash)
7. [Avoid Using `git reset`](#7-avoid-using-git-reset)
8. [Using `git log` to View Commit History](#8-using-git-log-to-view-commit-history)
9. **Advanced:**[Checking Out Files or Entire Commits with `git checkout`](#9-checking-out-files-or-entire-commits-with-git-checkout)
10. **Advanced:**[Rebasing Workflow](#10-rebasing-workflow)
11. **Advanced:**[Using Feature Branches](#11-using-feature-branches)

### 1. **Tracking Changes**

#### **How Git Tracks Your Changes**
After you modify files, Git won’t track them automatically. To include changes in the next commit, you need to **stage** them first using ```git add```.

#### **Key Commands for Staging and Committing**:
- **Stage a file**: ```git add <filename>```
- **Stage all modified files**: ```git add .```
- **Commit staged changes**: ```git commit -m "Your meaningful message"```

> **Tip**: Run ```git status``` frequently to see which files have been modified and staged. This helps avoid accidental commits.


### 2. **Writing Good Commit Messages**

Commit messages should describe *what* you changed and *why*. Clear, descriptive commit messages make it easy to understand the history of your project.

#### **Examples**:
- **Good**: ```git commit -m "Refactored login function for better readability"```
- **Bad**: ```git commit -m "fixed stuff"```

Always aim for clarity and precision when writing commit messages.


### 3. **Handling Merge Conflicts**

Merge conflicts happen when multiple people (or machines) modify the same file, and Git doesn’t know how to combine the changes.

#### **Steps to Resolve a Merge Conflict**:
1. **Git Marks the Conflict**: The conflicting areas in the file will be marked like this:
   ```plaintext
   <<<<<<< HEAD
   Your changes
   =======
   Remote changes
   >>>>>>> branch-name
   ```
2. **Edit the File Manually**: Choose which changes to keep, or combine both sets of changes. Remove the conflict markers (```^<<<```, ```===```, ```>>>```).
3. **Stage the Resolved File**: After resolving, stage the file:
   ```bash
   git add <filename>
   ```
4. **Commit the Merge**: Finalize the resolution by committing:
   ```bash
   git commit -m "Resolved merge conflict in <filename>"
   ```


### 4. **Using a `.gitignore` File**

#### **What is a `.gitignore`?**
A `.gitignore` file tells Git to ignore specific files or directories, preventing unnecessary or sensitive files from being tracked (e.g., log files, API keys).

#### **Example `.gitignore` for a Python Project**:
```plaintext
# Ignore Python cache files
__pycache__/
*.pyc

# Ignore environment files
.env

# Ignore log files
*.log
```

#### **How to Use It**:
- Create a `.gitignore` file in the root of your project.
- Add patterns for files and directories you want Git to ignore.

> **Important**: If a file is already tracked by Git, adding it to `.gitignore` will not stop Git from tracking it. You'll need to remove it from the history, which can be tricky.


### 5. **Undoing Mistakes**

#### **Unstaging Files (Accidental `git add`)**
If you accidentally staged a file, you can "unstage" it using:
```bash
git reset <filename>
```
This removes the file from the staged area without affecting the contents.

#### **Undoing a Commit Before Pushing**
If you want to undo your most recent commit (before pushing):
```bash
git reset --soft HEAD^
```
This will undo the last commit but keep your changes staged.

#### **Removing Sensitive Information (Seek Help for This)**
If you accidentally commit sensitive information like passwords or API keys, it can be very difficult to remove the data from the repository **if you’ve already pushed the commit**. 

#### **What to Do**:
1. **Remove the Sensitive Data**:
   - Delete the sensitive file or update it to remove the information.
   - Use ```git rm <filename> --cached``` if you need to remove the file from being tracked.

2. **Create a New Commit** to overwrite the sensitive information.

3. **Seek Help**: **If the information has already been pushed** to the remote repository, removing it is complex. Rewriting Git history requires special tools (e.g., ```git filter-repo```), and can affect others working on the project. **Ask for assistance** from your instructor or an experienced Git user before proceeding.

#### **Alternatively**:
If the project is small, or the history isn’t important, it might be easier to **delete the project on sc-gitlab and start fresh**. This will permanently remove the sensitive information from the Git history. You can recreate the project, initialize a new Git repository, and push the current state of the project again—sacrificing the project history but ensuring sensitive information is removed.


### 6. **Saving Work Temporarily with `git stash`**

#### **Why Use `git stash`?**
If you're working on something and need to switch tasks (like pulling new changes), but you're not ready to commit, `git stash` lets you **save your uncommitted changes temporarily**.

#### **Example Workflow Using `git stash`**:
1. **Stash Your Changes**:
   ```bash
   git stash
   ```
   This command removes your local changes but stores them safely.

2. **Pull New Changes**:
   ```bash
   git pull
   ```
   Now you can pull the latest updates without committing your unfinished work.

3. **Reapply Your Stashed Changes**:
   ```bash
   git stash pop
   ```
   Your previously stashed changes are reapplied on top of the pulled changes.

> **Tip**: `git stash` is a life-saver when you need to handle tasks like resolving merge conflicts without committing incomplete work.


### 7. **Avoid Using `git reset`**

#### **Why `git reset` Can Be Dangerous**:
`git reset` can undo changes to your files, commits, or even entire Git history. If you're not careful, you might lose important work that is difficult to recover.

#### **Safer Alternatives**:
- **Use `git revert`**: If you need to undo a specific commit, ```git revert``` is a safer option. It creates a new commit that undoes the previous one, without rewriting history:
  ```bash
  git revert <commit_id>
  ```
- **Use `git stash`**: Temporarily save uncommitted changes without resetting or committing them.


### 8. **Using `git log` to View Commit History**

#### **What is `git log`?**
`git log` is a command that displays the history of commits in your repository. It’s a useful way to see what changes were made, when they were made, and who made them. Understanding your commit history can help you track down changes, debug issues, or simply review your project’s progress.

#### **Basic Usage of `git log`**
The simplest way to view your commit history is to run:
```bash
git log
```
This will show a list of commits in reverse chronological order (most recent first). For each commit, you’ll see:
- **Commit hash**: A unique identifier for each commit (e.g., `a3c9b5d`).
- **Author**: Who made the commit.
- **Date**: When the commit was made.
- **Commit message**: The message associated with the commit.

#### **Common Options for `git log`**
```git log``` has many options to customize the output. Here are a few useful ones:

1. **Display a Compact One-Line Log**:
   If you want to see a shorter, more compact version of the log with just one line per commit:
   ```bash
   git log --oneline
   ```
   This will show each commit’s hash and message in a single line. Example:
   ```plaintext
   a3c9b5d Refactored login function
   7c1f5b2 Fixed bug in user signup
   ```

2. **Show a Limited Number of Commits**:
   If you're only interested in the most recent commits, you can limit the output:
   ```bash
   git log -n 5
   ```
   This will display only the last 5 commits. You can change the number to display as many or as few as you want.

3. **Display Commits for a Specific File**:
   To view the commit history of a particular file, you can specify the file name:
   ```bash
   git log <filename>
   ```
   This is useful when you want to track changes made to a specific file.

4. **Show Detailed Changes for Each Commit**:
   If you want to see the actual changes (a "diff") introduced in each commit, you can use the `-p` option:
   ```bash
   git log -p
   ```
   This will display the differences introduced in each commit, showing exactly what lines were added, removed, or modified. This is helpful when you want to review the specific changes in detail.

5. **Format the Output**:
   You can customize the format of the log to include specific information:
   ```bash
   git log --pretty=format:"%h - %an, %ar : %s"
   ```
   - %h: Short commit hash.
   - %an: Author name.
   - %ar: Relative date (e.g., 2 days ago).
   - %s: Commit message.
   
   Example output:
   ```plaintext
   a3c9b5d - John Doe, 3 days ago : Refactored login function
   7c1f5b2 - Jane Smith, 5 days ago : Fixed bug in user signup
   ```

6. **Show Commit History as a Graph**:
   You can visualize the commit history as a branching graph, which is useful when working with branches:
   ```bash
   git log --graph --oneline
   ```
   This will show the log as a simple ASCII graph that represents the branching structure of your commits.

#### **Searching the Commit Log**
If you're looking for a specific commit or keyword in the log (e.g., a bug fix), you can search the commit messages:
```bash
git log --grep="<keyword>"
```
Replace ```<keyword>``` with the term you’re searching for. This will only show commits whose messages contain the specified keyword.


### 9. **Checking Out Files or Entire Commits with `git checkout`**
**Advanced**

#### **Checking Out a File from a Previous Commit**
Sometimes you may want to restore a file to the state it was in a specific commit without affecting the rest of your working directory. You can use `git checkout` to do this.

1. **Find the commit hash**: Use `git log` to locate the commit you want to check out.
2. **Check out the file from that commit**:
   ```bash
   git checkout <commit_hash> -- <filename>
   ```
   This restores the file to how it was in the specified commit. Your working directory will now contain this version of the file, but your other files will remain unchanged.

Example:
```bash
git checkout a3c9b5d -- script.py
```
This restores `script.py` to its state in commit `a3c9b5d`.

#### **Checking Out a Specific Commit**
If you want to revert your entire project to a particular commit, you can check out the whole repository at that commit:
```bash
git checkout <commit_hash>
```
This will move your working directory to the state it was in at that specific commit. This puts you in a **detached HEAD state**, which means you're not on any branch. Any changes made here will not belong to any branch unless you explicitly create one.

To safely return to your latest work after viewing a previous commit, switch back to your branch with:
```bash
git checkout main
```
> **Note**: You don’t need to worry about branching yet, but later on, you’ll learn that creating a new branch before checking out a specific commit can help you explore older versions without losing track of your main work.

### Bonus: **Introduction to `git rebase`**

#### **What is `git rebase`?**
```git rebase``` is a powerful tool in Git that is used to **rewrite commit history**. When you rebase, you take a series of commits from one branch and apply them on top of another branch. It’s often used to clean up the history or incorporate changes from one branch into another in a linear, sequential way.

#### **How `git rebase` Works**
Imagine you have two branches, `main` and `feature`. You made several commits on the `feature` branch, but the `main` branch has new changes that you want to integrate into your work. Instead of merging (which would create a merge commit), you can use `git rebase` to **apply your commits on top of `main`** in a way that looks like they were created after the new changes in `main`.

### 10. **Rebasing Workflow**:
1. First, switch to the branch you want to rebase:
   ```bash
   git checkout feature
   ```
2. Then, start the rebase:
   ```bash
   git rebase main
   ```
   This tells Git to take the changes from the `feature` branch and apply them on top of the `main` branch.

3. If there are conflicts during the rebase, Git will stop and allow you to resolve the conflicts. Once resolved, continue the rebase with:
   ```bash
   git rebase --continue
   ```

#### **Rebase vs. Merge**
- **Merge**: Combines branches and creates a **merge commit** that shows the history of both branches. This can lead to a non-linear history.
- **Rebase**: Takes the changes from one branch and applies them on top of another, creating a **linear history** without a merge commit. This can make the history cleaner.

#### **When to Use `git rebase`**
- Use rebase when you want to **keep your commit history clean** and avoid the clutter of merge commits.
- Avoid using rebase on public branches that others are working on, because it rewrites history and can make collaboration confusing.


## 11. **Using Feature Branches**
**Advanced**

#### **What Are Feature Branches?**
A **feature branch** is a Git practice where you create a separate branch from the main development line (typically `main` or `master`) to work on a new feature, bug fix, or any isolated task. Feature branches allow you to keep the main branch clean and stable while developing and testing new functionality.

This branching model is widely used in teams to ensure that the work on new features is kept separate from the stable codebase until the feature is fully tested and ready to be integrated.

#### **Why Use Feature Branches?**
Feature branches offer several benefits:
- **Isolation**: Keep experimental or incomplete work separate from the main codebase.
- **Collaboration**: Multiple developers can work on different features simultaneously without stepping on each other’s toes.
- **Safe integration**: You can merge changes only when the feature is complete, tested, and stable.
- **Easy rollback**: If a feature doesn’t work as expected, you can easily discard the branch without affecting the main project.


### **Creating and Using Feature Branches**

#### **Step 1: Create a New Feature Branch**
To start working on a new feature, you first need to create a new branch. This branch will be based on the latest state of the `main` branch (or any branch you want to start from).

1. **Switch to the `main` branch**:
   ```bash
   git checkout main
   ```
   This ensures your feature branch will be based on the latest version of `main`.

2. **Pull the latest changes from the remote** (if working in a shared repository):
   ```bash
   git pull origin main
   ```
   This ensures you have the most up-to-date code.

3. **Create a new branch**:
   ```bash
   git checkout -b feature/new-feature
   ```
   This creates a new branch called `feature/new-feature` and switches to it. The `feature/` prefix is optional, but it's a good naming convention to distinguish feature branches from other types of branches (like `bugfix/` or `hotfix/`).

#### **Step 2: Work on Your Feature**
Now that you’re on the feature branch, you can begin making changes. As you make progress, commit your changes regularly:
```bash
git add <file>
git commit -m "Added new component for user authentication"
```

> **Tip**: Use clear, descriptive commit messages that explain *what* you did and *why*.


### **Integrating Your Feature Branch**

Once your feature is complete and fully tested, you’ll want to integrate it back into the `main` branch. Before doing so, make sure your feature branch is up to date with any new changes from `main` to avoid conflicts.

#### **Step 3: Keep Your Feature Branch Updated**
To avoid potential merge conflicts when you finally merge your feature branch into `main`, regularly update your feature branch with any changes from `main`. You can do this by rebasing or merging.

1. **Merge the latest `main` into your feature branch**:
   ```bash
   git checkout feature/new-feature
   git pull origin main
   git merge main
   ```
   This pulls the latest changes from `main` and merges them into your feature branch.

   If there are conflicts, Git will notify you, and you’ll need to resolve them manually. Once resolved, commit the merged changes:
   ```bash
   git add <conflicted_file>
   git commit -m "Resolved merge conflicts between main and feature/new-feature"
   ```

   Alternatively, you can use **rebasing** to keep a clean, linear history. Rebasing re-applies your commits on top of the latest version of `main`:
   ```bash
   git rebase main
   ```
   After rebasing, you might also need to resolve conflicts.

> **Note**: Rebasing rewrites the history, so it’s best not to rebase public branches shared with other developers.


### **Step 4: Merge the Feature Branch into `main`**

Once your feature branch is up to date and your changes are ready, you can merge the feature branch into `main`.

1. **Switch to the `main` branch**:
   ```bash
   git checkout main
   ```

2. **Merge your feature branch**:
   ```bash
   git merge feature/new-feature
   ```
   This integrates the feature branch into `main`. If there are any conflicts, resolve them as needed and commit the result.

3. **Push the merged changes to the remote**:
   ```bash
   git push origin main
   ```
   This pushes the updated `main` branch to the remote repository, making the new feature available to everyone else working on the project.

#### **Step 5: Delete the Feature Branch (Optional)**
After merging, the feature branch is no longer needed, so you can safely delete it:
```bash
git branch -d feature/new-feature
```
If the branch has already been pushed to the remote repository, delete it from there as well:
```bash
git push origin --delete feature/new-feature
```

> **Tip**: Deleting feature branches helps keep your repository clean and focused only on active work.


### **When to Use Feature Branches**

Feature branches are useful in the following scenarios:
- **Developing new features**: Keep incomplete or experimental features out of the `main` branch until they are fully developed and tested.
- **Fixing bugs**: You can create a `bugfix/` branch to work on fixes independently, without affecting ongoing feature development.
- **Collaborative work**: When working in a team, each member can work on their own feature branch, and their work can be merged into `main` only when it's complete and conflict-free.

