# Common Git + SSH Scenarios and Solutions for Windows Users (Git Bash / WSL)

See also: https://antonz.org/git-by-example/
---

## 1. SSH Key Not Generated Yet / Public Key Not Added to sc-gitlab.ufv.ca

**Scenario**: The user hasn't generated an SSH key or added it to their GitLab profile.

### Solution:

1. **Open Git Bash** or **WSL**.
   
2. **Check if an SSH key exists**:
   ```bash
   ls -la ~/.ssh
   ```
   If you don’t see a file named `id_rsa.pub`, follow the next steps to generate a new SSH key.

3. **Generate a new SSH key**:
   Run the following command:
   ```bash
   ssh-keygen
   ```
   - Press **Enter** to accept the default file location (`/c/Users/your-username/.ssh/id_rsa` for Git Bash).
   - Press **Enter** to skip setting a passphrase (or set one if you prefer).
   
4. **View and copy the public key**:
   Run the following command to view your public key:
   ```bash
   cat ~/.ssh/id_rsa.pub
   ```

   **Example output**:
   ```
   ssh-rsa AAAAB3NzaC1yc2EAAAABIwAAAQEArzdK5mKhTR...tOvzjygS4oB6E user@machine
   ```

5. **Add the public key to GitLab**:
   - Copy the entire public key.
   - Go to `sc-gitlab.ufv.ca` and log in.
   - Navigate to **Preferences** → **SSH Keys**.
   - Paste the public key into the field and click **Add Key**.

---

## 2. Server Key Not in Known Hosts File During `git clone`

**Scenario**: The user runs `git clone` to clone a repository from `sc-gitlab.ufv.ca`, but the server's key is not in the known hosts file. Instead of typing **y** to trust the host, the user keeps pressing **Enter**.

### Solution:

1. **Run the `git clone` command again**:
   ```bash
   git clone git@sc-gitlab.ufv.ca:username/repository.git
   ```

2. When prompted with:
   ```
   The authenticity of host 'sc-gitlab.ufv.ca (IP)' can't be established.
   Are you sure you want to continue connecting (yes/no)?
   ```
   **Type** `yes` and press **Enter**. This will add the server's key to the `known_hosts` file.

---

## 3. Cloning a Repository for the First Time

**Scenario**: The user needs to clone a repository using SSH and understand potential issues.

### Solution:

1. **Find the SSH URL**:
   - On GitLab, navigate to your repository.
   - Click **Clone** and copy the **SSH URL** (e.g., `git@sc-gitlab.ufv.ca:username/repository.git`).

2. **Clone the repository**:
   Run the command in your terminal:
   ```bash
   git clone git@sc-gitlab.ufv.ca:username/repository.git
   ```

3. **Potential errors**:
   - **Access Denied**: The SSH key has not been added to GitLab. See solution 1.
   - **Repository not found**: Double-check the SSH URL and ensure you have access.

---

## 4. Changing to the Directory with the Repository

**Scenario**: After cloning, the user needs to move to the correct directory and understand basic terminal commands.

### Solution:

1. **Check your current directory**:
   ```bash
   pwd
   ```
   **Example output**:
   ```
   /c/Users/your-username
   ```

2. **List all files and directories (detailed)**:
   ```bash
   ls -la
   ```
   **Example output**:
   ```
   drwxr-xr-x 3 your-username your-username 4096 Oct  1 12:34 .
   drwxr-xr-x 5 your-username your-username 4096 Oct  1 11:00 ..
   drwxr-xr-x 8 your-username your-username 4096 Oct  1 12:34 repository
   ```

3. **Change to the repository directory**:
   ```bash
   cd repository
   ```

4. **Confirm you are in the correct directory**:
   ```bash
   pwd
   ```
   **Example output**:
   ```
   /c/Users/your-username/repository
   ```

---

## 5. Navigating the Directory Structure and Finding Your Project Directory

**Scenario**: The user needs to understand basic navigation, how to find and enter their project directory, and the importance of pulling the latest changes.

### Solution:

1. **List files and directories with details**:
   ```bash
   ls -la
   ```
   **Example output**:
   ```
   drwxr-xr-x 3 your-username your-username 4096 Oct  1 12:34 .
   drwxr-xr-x 5 your-username your-username 4096 Oct  1 11:00 ..
   -rw-r--r-- 1 your-username your-username    0 Oct  1 12:35 newfile.txt
   drwxr-xr-x 2 your-username your-username 4096 Oct  1 12:34 directory
   ```

2. **Tab completion for navigating directories**:
   - As you type a directory name, you can press **Tab** to automatically complete the directory or file name.
   - For example, if the directory name is `my_project`, you can type:
     ```bash
     cd my_pro
     ```
     Then press **Tab**, and it will autocomplete to:
     ```bash
     cd my_project
     ```

3. **Find and enter your project directory**:
   ```bash
   cd my_project_directory
   ```

4. **Confirm the current directory**:
   ```bash
   pwd
   ```
   **Example output**:
   ```
   /c/Users/your-username/my_project_directory
   ```

5. **Reminder: Always pull the latest changes**:
   Before you start working, **always pull the latest changes** from the remote repository:
   ```bash
   git pull
   ```

---

## 6. Using `nano` to Edit Files

**Scenario**: The user needs to edit files in the terminal using `nano`.

### Solution:

1. **Open a file in `nano`**:
   ```bash
   nano filename.txt
   ```

2. **Edit the file**.

3. **Save the file**:
   - Press **Ctrl + O**, then **Enter** to save.
   
4. **Exit `nano`**:
   - Press **Ctrl + X** to exit.

---

## 7. Standard Git Workflow

**Scenario**: User needs to follow the typical workflow of pulling, editing, committing, and pushing changes.

### Solution:

1. **Pull the latest changes from the remote repository**:
   ```bash
   git pull
   ```

2. **Edit files** (e.g., using `nano`).

3. **Stage changes**:
   ```bash
   git add filename.txt
   ```

4. **Commit changes**:
   ```bash
   git commit -m "Your commit message"
   ```

5. **Push changes to GitLab**:
   ```bash
   git push
   ```

---

## 8. Common Issues When Pushing to Git

**Scenario**: Problems occur when trying to push changes to the remote repository.

### Solution:

1. **Remote contains changes not in your local repository**:
   **Cause**: Someone else has made changes to the remote repository that are not reflected in your local copy.
   **Error message**:
   ```
   error: failed to push some refs to 'git@sc-gitlab.ufv.ca:username/repository.git'
   ```
   **Solution**:
   - Pull the latest changes using rebase:
     ```bash
     git pull --rebase
     ```

2. **Incompatible histories (Avoid merging unrelated histories)**:
   **Cause**: The local and remote repositories have different histories.
   **Error message**:
   ```
   fatal: refusing to merge unrelated histories
   ```
   **Solution**:
   Avoid merging unrelated histories. Instead:
   - Ensure you are working with the correct repository.
   - Delete your local repository and clone it again using the correct SSH URL.

3. **Merge conflicts**:
   **Cause**: Conflicting changes in the same file.
   **Error message**:
   ```
   CONFLICT (content): Merge conflict in filename.txt
   ```
   **Solution**:
   - Open the conflicted file, resolve the conflicts, then:
     ```bash
     git add filename.txt
     git commit
     git push
     ```

4. **Vim editor opens during merge**:
   **Cause**: Git defaults to `vim` for editing commit messages.
   **Solution**:
   - To save and exit `vim`, press **Esc**, then type `:wq` and press **Enter**.

---

## 9. Default Branch Not Set

**Scenario**: The default branch has not been set in the repository.

### Solution:

1. **Set the default branch (main)**:
   ```bash
   git branch -M main
   ```

2. **Push the main branch**:
   ```bash
   git push -u origin main
   ```

---

## 10. Merge Mode Not Set

**Scenario**: The repository doesn’t have a configured merge mode, causing problems during merges.

### Solution:

1. **Check the merge mode**:
   ```bash
   git config --global pull.rebase
   ```

2. **Set rebase as the default merge mode**:
   ```bash
   git config --global pull.rebase true
   ```

---

## 11. User Name and Email Not Set

**Scenario**: User hasn’t configured their Git username and email, which can lead to issues when committing changes.

### Solution:

1. **Check if `user.name` and `user.email` are set**:
   Run the following commands:
   ```bash
   git config --global user.name
   git config --global user.email
   ```

2. **Set the username and email**:
   If they are not set, configure them like this:
   ```bash
   git config --global user.name "Your Name"
   git config --global user.email "your_email@example.com"
   ```


## 12. Need to delete directory

When deleting directories, it is important to delete recursively:

```bash
rm -r directoryname

```

