# Introduction

Python is an object-oriented programming (OOP) language, which means it
allows us to model real-world entities using *objects* and *classes*.
Understanding objects and classes is fundamental to working with Python,
especially as programs grow in complexity.

# What is a Class?

A **class** in Python is a blueprint or template for creating objects.
It defines a set of attributes (data) and methods (functions) that the
objects created from the class will have. Think of a class as a
blueprint for constructing houses; each house you build from that
blueprint is an object.

In Python, you define a class using the `class` keyword.

## Example 

```python

    class Dog:
        # This is the blueprint for a Dog object

        # Constructor method
        def __init__(self, name, breed):
            # Attributes
            self.name = name
            self.breed = breed

        # Method: a function belonging to the class
        def bark(self):
            print(f"{self.name} says: Woof!")

```

**Key Points:**

-   `__init__(self, ...)`: The **constructor method** called when you
    create a new object (instance). It initializes the object's
    attributes.

-   `self`: Refers to the specific instance of the class (i.e., the
    object being created or manipulated).

-   `name`, `breed`: **Attributes** that store data specific to each
    instance (e.g., each dog will have its own name and breed).

-   `bark(self)`: A **method** that defines behavior for instances of
    the class.

# What is an Object?

An **object** is an instance of a class. While a class defines the
structure and behavior, an object is the actual entity created based on
the class. Each object can have different values for its attributes, but
all objects of the same class will share the same methods.

## Example 

```python
    # Create an object (instance) of the Dog class
    dog1 = Dog("Buddy", "Golden Retriever")
    dog2 = Dog("Max", "Bulldog")

    # Access attributes
    print(dog1.name)  # Output: Buddy
    print(dog2.breed)  # Output: Bulldog

    # Call methods
    dog1.bark()  # Output: Buddy says: Woof!
```

In this example:

-   `dog1` and `dog2` are **objects** (or instances) of the `Dog` class.

-   Each object has its own data (i.e., name and breed) but shares the
    `bark()` method from the `Dog` class.

# Instance vs. Class Variables

Python classes can have two types of variables:

-   **Instance variables**: Defined within methods and belong to each
    object instance. These are specific to each object (e.g., `name` and
    `breed`). Referenced within the class as `self.attribute` and outside
    the class as `instancename.attribute`

-   **Class variables**: Shared across all instances of the class. They
    are defined directly inside the class, but outside of any methods.
    Referenced within the class and outside the class as `Classname.attribute`

## Example 

```python

    class Dog:
        species = "Canis lupus"  # Class variable shared by all instances
        
        def __init__(self, name, breed):
            self.name = name       # Instance variable
            self.breed = breed     # Instance variable

    dog1 = Dog("Buddy", "Golden Retriever")
    dog2 = Dog("Max", "Bulldog")

    # Access class variable
    print(dog1.species)  # Output: Canis lupus
    print(dog2.species)  # Output: Canis lupus
```

Here, `species` is a class variable that will be the same for all
instances of `Dog`.

To change species:
```python
  Dog.species = "Canis lupus superioris"
```
All dog instances, including past creations automatically get the new species attribute


# Methods in Classes

There are three types of methods commonly used in Python classes:

-   **Instance Methods**: Take `self` as the first parameter and can
    access instance variables. For COMP150 we will only use these.

-   **Class Methods**: (SKIP for COMP150) Take `cls` as the first parameter and work with
    class variables. Use the `@classmethod` decorator. (not used for COMP150)

-   **Static Methods**:(SKIP for COMP150) Do not take `self` or `cls` and are defined
    using the `@staticmethod` decorator. (likely not going to be applicable for COMP150)

# Encapsulation and Private Variables

In Python, encapsulation refers to restricting access to methods and
variables. By convention, variables prefixed with an underscore (e.g.,
`_variable`) are treated as \"private.\" This is not enforced by Python,
but it is a common practice. Another way to look at this is to say that
the implement is \"unstable\" and can be changed, so should not be used
by other classes directly.

To make a variable more strongly private, you can use double underscores
(`__`), which invoke name mangling. Python changes the variable name
internally to avoid accidental access. Access to `object.__function__`
will fail, but name mangling will make it possible to still access 
using something like `object._Classname__function__`

## Example 

```python

    class Car:
        def __init__(self, brand):
            self.__brand = brand  # Private variable

        def set_brand(self, brand):
            if self.__brand == "Chevrolet" and brand == "Pontiac":
              self.__brand = brand

        def get_brand(self):
            return self.__brand

    car = Car("Toyota")
    print(car.get_brand())  # Output: Toyota
    print(car.__brand)  # Raises an AttributeError
    car.set_brand("Lexus")  # has no effect
```


# Example

```
    class Balance:
        def __init__(self, amount):
            if amount < 0:
              self.__amount = 0
            else:
              self.__amount = amount # Private variable

        def deposit(self, amount):
            if amount < 0:
              raise ValueError("cannot deposit negative") # see session4 for raising errors
            # takes a positive number indicating how much to deposit
            self.__amount = self.__amount + amount
            # returns amount deposited
            return amount

        def withdraw(self, amount):
            # takes a positive number indicating how much to withdraw 
            # returns amount withdrawn, if any
            if self.__amount - amount < 0:
              return 0

            self.__amount = self.__amount - amount
            return amount

    balance = Balance(5)
    x = balance.withdraw(10)
    # what value is x? (zero)
```

# Inheritance

Inheritance allows one class (the child or subclass) to inherit
attributes and methods from another class (the parent or superclass).
This promotes code reuse.

See also: mixins (or in C++ multiple inheritance)

(For COMP150, cursory coverage only)

## Example 

```python
    class Mammal:
        def __init__(self, name):
            self.name = name
        
        def speak(self):
            print(f"{self.name} makes a sound")

        def breathe(self):
            print(f"{self.name} breathes")

    class Dog(Mammal):
        def speak(self):
            print(f"{self.name} barks")

    dog = Dog("Buddy")
    dog.speak()  # Output: Buddy barks
    dog.breathe() # Output: Buddy breathes

```

Here, `Dog` is a subclass of `Mammal`. It inherits the `__init__` method
and the `breathe` method
from `Mammal`, but overrides the `speak()` method.


# References to Standard Python Documentation

-   Official Python documentation on
    [Classes](https://docs.python.org/3/tutorial/classes.html): This
    section covers the concept of classes, objects, inheritance, and
    more.

-   Documentation on
    [OOP](https://docs.python.org/3/glossary.html#term-object-oriented):
    Definitions and examples of object-oriented programming terms.

-   [PEP 8](https://peps.python.org/pep-0008/): Style guide for Python,
    which includes guidelines for naming classes and instance variables.

# Review

-   **Classes** define the structure and behavior of objects.

-   **Objects** are instances of classes.

-   Attributes represent data, while methods define behavior.

-   Concepts such as inheritance, encapsulation, and class/instance
    variables allow for more complex and organized code.

