## Virtual Environments and Package Management in Python

When developing Python applications, managing dependencies and isolating
project environments is crucial. Different projects often require
different versions of the same libraries, so managing them globally can
lead to conflicts. This is where **virtual environments** and **package
management** come in handy.

### What is a Virtual Environment?

A **virtual environment** is an isolated environment that contains its
own installation of Python and its own set of libraries. It allows you
to:

-   Avoid dependency conflicts between projects.

-   Work on different projects with different versions of libraries.

-   Manage project-specific dependencies without affecting the global
    Python environment.

#### Setting Up a Virtual Environment

To create and manage virtual environments, Python's `venv` module is
commonly used. Here's how you can create a virtual environment:

1.  **Create a virtual environment**:

    ```bash 
    python3 -m venv env
    ```

    This creates a directory named `env` containing a standalone Python
    installation.

  **NB:** You will want to add the `env` directory to your `.gitignore` file,
    and add the `.gitignore` file to your git repository, to avoid accidentally
    adding the `env` directory to your tracked files.
    ```bash
    echo "env" >> .gitignore
    git add .gitignore
    git commit -m "added gitignore entry to not track the env directory"
    git push
    ```


2.  **Activate the virtual environment**:

    -   On **Linux/macOS**:

        ```bash 
        source env/bin/activate
        ```

    -   On **Windows**:

        ```
        .\env\Scripts\activate
        ```

    After activation, the terminal prompt will reflect the virtual
    environment.

3.  **Deactivate** the virtual environment when you're done:

    ```bash
    deactivate
    ```

### Package Management with `pip`

Python uses the package manager **`pip`** to install and manage
third-party libraries. When working inside a virtual environment, `pip`
installs packages in that environment only, preventing conflicts with
global installations.

#### Example: Using the `requests` Module

Let's use the popular **`requests`** library as an example.

1.  **Install the `requests` module** using `pip` inside your virtual
    environment:

    ```bash
    pip install requests
    ```

2.  **Check the installed packages**:

    ```bash
    pip list
    ```

3.  **Using the `requests` module** in a Python script:

    ```python
    import requests

    response = requests.get('https://api.github.com')
    print(response.status_code)
    print(response.json())
    ```

4.  **Freeze dependencies** into a `requirements.txt` file:

    ```bash
    pip freeze > requirements.txt
    ```

5.  **Install dependencies** from a `requirements.txt` file:

    ```bash
    pip install -r requirements.txt
    ```

### Managing Virtual Environments with Miniconda

**Miniconda** is another tool for managing environments and packages,
particularly popular in data science. Unlike `venv` and `pip`, Miniconda
uses the **`conda`** package manager, which provides a more
comprehensive environment and dependency management system.

## Differences Between `venv`/`pip` and Miniconda

  **Feature**                  `venv` + `pip`                        `Miniconda` (`conda`)
  ---------------------------- ------------------------------------- ---------------------------------------------------
  **Package Sources**          PyPI                                  PyPI + Conda repositories
  **Binary Package Support**   Limited (may require compiling)       Excellent (pre-built binaries for many libraries)
  **Cross-Language Support**   Python only                           Multiple languages (Python, R, etc.)
  **Ease of Use**              Simple and lightweight                More complex, but powerful for data science
  **Environment Management**   Basic, isolated Python environments   Comprehensive, with more features

## Example: Installing `requests` with Miniconda

1.  **Create a conda environment**:

    ```bash
    conda create --name myenv python=3.10
    ```

2.  **Activate the environment**:

    ```bash
    conda activate myenv
    ```

3.  **Install `requests`**:

    ```bash
    conda install requests
    ```

4.  **List installed packages**:

    ```bash
    conda list
    ```

### General Guidance on Virtual Environments

For most basic Python projects, **`venv`** + **`pip`** is fine.
However,  **Miniconda** offers more advanced environment
management and better handling of dependencies.


## Extended discussion of functions

See: session4b
