# Basic Function Structure

**NB:** This document was created as an extended answer around functions and type hints, as seen in Assignment 2

A function is a reusable block of code that performs a specific task. It
is defined using the `def` keyword, followed by the function name and
parentheses `()`.

```python
def myfunction():
  pass
```

`pass` is a placeholder, meaning \"do nothing.\" It's often used in
function definitions to create a function that doesn't do anything yet
but can be filled in later.

# Functions with Mandatory Positional Parameters

A function can accept **positional parameters**, which are arguments you
must provide when calling the function.

```python
def greet(name):
  print(f"Hello {name}")
```

The function `greet()` takes a **single mandatory parameter** `name`. If
you call `greet()` without providing a value (e.g., `greet()`), Python
raises an error because `name` is required.

**Example Call:**

```python
greet("Mo")  # Output: Hello Mo
```

# Functions with Multiple Mandatory Positional Parameters

You can define functions that take multiple parameters.

```python
def multigreet(name, n):
  for i in range(n):
    print(f"Hello {name}")
```

`multigreet()` takes two parameters: `name` and `n`. It repeats the
greeting `n` times using a `for` loop.

**Example Call:**

```python
multigreet("Mo", 3)
# Output:
# Hello Mo
# Hello Mo
# Hello Mo
```

# Functions with an Arbitrary Number of Positional Arguments

Sometimes, you may want to pass a variable number of arguments to a
function. You can use the `*` operator to collect additional arguments
into a tuple.

```python
def myprint(*words):
  for word in words:
    print(word)
  print(" ".join(words))  # Joins the words into a single string
```

`*words` collects all the arguments passed into the function into a
tuple.

**Example Call:**

```python
myprint("Joe", "Raj", "Mo")
# Output:
# Joe
# Raj
# Mo
# Joe Raj Mo
```

# Positional and Optional Parameters

You can combine **mandatory positional parameters** and **optional
arbitrary arguments** in the same function.

```python
def myprint(firstword, *morewords):
  pass
```

In this example, `firstword` is mandatory, while `*morewords` can
collect additional arguments if provided.

# Keyword Arguments with Default Values

Functions can have **keyword arguments** with default values. This
allows for flexible function calls where certain arguments are optional.

```python
def myprintkw(name="", n=0):
  for i in range(n):
    print(name)
```

`name` and `n` are keyword arguments with default values of an empty
string and `0`.

**Example Call:**

```python
myprintkw(name="Ash", n=9)
# Output: Ash (printed 9 times)

myprintkw(n=9, name="Ash")
# Output: Ash (printed 9 times)
```
```

# Functions with Arbitrary Keyword Arguments (`**kwargs`)

You can accept arbitrary keyword arguments using `**kwargs`. These are
collected into a dictionary.

```python
def myprintkwext(name="", n=0, **kwargs):
  for i in range(n):
    print(name)
  print(kwargs)  # Prints extra keyword arguments as a dictionary
```

**Example Call:**

```python
myprintkwext(name="Joe", n=2, extra="value", another="keyword")
# Output:
# Joe
# Joe
# {'extra': 'value', 'another': 'keyword'}
```

# Combining Positional, Optional, and Keyword Arguments

You can create functions that combine positional parameters, optional
arbitrary arguments, and arbitrary keyword arguments.

```python
def myprintkwextopt(name="", n=0, *args, **kwargs):
  for i in range(n):
    print(name)
  print(kwargs)
```

See also: [Demystifying args and kwargs](https://realpython.com/python-kwargs-and-args/)

# Type Hints (Advanced)

Type hints allow you to specify the expected data types of function
parameters and return values. This can be useful for code readability
and static type checking, but it is not enforced at runtime.

# see pydoc javadoc, doctest for intermingling code and comments and documents

```python
def greet(name: str) -> None:
  print(f"Hello {name}")
```

`name: str` specifies that `name` should be a string, and `-> None`
indicates that the function does not return a value.

**Further Reading:** For more information on type hints, refer to
Python's documentation: <https://docs.python.org/3/library/typing.html>

# Raising Errors

You can raise exceptions using the `raise` keyword. This is useful for
signaling that something has gone wrong.

```python
x = 0
if x < 100:
  raise ValueError("number is too low")
```

This code raises a `ValueError` if `x` is less than 100.

# Future Topics for Exploration

-   **Handling Exceptions**: Techniques for catching and managing errors
    using `try`, `except`, and `finally`.

-   **Asyncio Functions**: Functions that allow for asynchronous
    programming, enabling concurrent tasks.

-   **Test Driven Development**: Writing tests before code to ensure
    your program functions as expected.

-   **2D Graphics frameworks**: Writing code with a graphical target

