# Basic File I/O

In Python, file input and output (I/O) allows us to read from or write
to files on disk. Python provides built-in functions such as `open()`,
`read()`, `write()`, and `close()` for handling files.

## Opening a File

To open a file, use the `open()` function, which takes the file path and
mode (read, write, append, etc.) as parameters.

```python
# Syntax for opening a file
file = open('filename.txt', 'mode')
```

Common file modes include:

-   `’r’` -- Read (default mode)

-   `’w’` -- Write (overwrites if the file exists)

-   `’a’` -- Append (adds to the end of the file)

-   `’b’` -- Binary mode (for non-text files)

[Reference: Python File Handling -
`open()`](https://docs.python.org/3/library/functions.html#open)

## Reading a File

To read a file, use `read()`, `readline()`, or `readlines()`.

```python
# Reading the entire file content
with open('example.txt', 'r') as file:
    content = file.read()
    print(content)

# Reading line by line
with open('example.txt', 'r') as file:
    for line in file:
        print(line.strip())  # removes newline character
```

[Reference: Reading and Writing
Files](https://docs.python.org/3/tutorial/inputoutput.html#reading-and-writing-files)

## Writing to a File

To write to a file, use the `write()` method.

```python
# Writing to a file
with open('output.txt', 'w') as file:
    file.write("Hello, world!\n")

# Appending to a file
with open('output.txt', 'a') as file:
    file.write("This is appended text.\n")
```

[Reference: Real Python - Reading and Writing Files in
Python](https://realpython.com/read-write-files-python/)

# Console Input and Output

## Console Output (`print`)

To display information on the console, use the `print()` function.

```python
print("Hello, world!")

# Printing multiple items
name = "Alice"
age = 25
print("Name:", name, "Age:", age)

# Using formatted strings
print(f"Name: {name}, Age: {age}")
```

[Reference: Python Built-in Function
`print()`](https://docs.python.org/3/library/functions.html#print)

## Console Input (`input`)

To accept user input, use the `input()` function, which always returns a
string.

```python
# Asking the user for input
name = input("Enter your name: ")
print(f"Hello, {name}!")

# Converting input to a different type
age = int(input("Enter your age: "))
print(f"You are {age} years old.")
```

[Reference: Python Built-in Function
`input()`](https://docs.python.org/3/library/functions.html#input)

# Serialization to and from JSON Files

Python's `json` module allows you to serialize (convert Python objects
to JSON format) and deserialize (convert JSON back into Python objects).

[JSON standard](https://www.json.org/json-en.html)

## Writing to a JSON File

To serialize Python objects to a JSON file, use the `json.dump()`
function.

```python
import json


# Example data
data = {
    "name": "Alice",
    "age": 25,
    "is_student": True
}

# Write data to a JSON file
with open('data.json', 'w') as file:
    json.dump(data, file)
```

[Reference: Writing JSON to
Files](https://docs.python.org/3/library/json.html#json.dump)

## Reading from a JSON File

To deserialize JSON data from a file, use `json.load()`.

```python
import json

# Read data from a JSON file
with open('data.json', 'r') as file:
    data = json.load(file)
    print(data)  # Output: {'name': 'Alice', 'age': 25, 'is_student': True}
```

[Reference: Reading JSON from
Files](https://docs.python.org/3/library/json.html#json.load)

## Serializing to JSON String

To serialize Python objects to a JSON string, use `json.dumps()`.

```python
json_string = json.dumps(data)
print(json_string)
```

## Deserializing from JSON String

To deserialize JSON strings back into Python objects, use
`json.loads()`.

```python
data = json.loads(json_string)
print(data)
```

[Reference: Real Python - Working with JSON Data in
Python](https://realpython.com/python-json/)

# Serialization to and from CSV Files

Python's `csv` module handles reading and writing CSV (Comma-Separated
Values) files.

## Writing to a CSV File

To write data to a CSV file, use `csv.writer()`.

```python
import csv

# Data to write
rows = [
    ['Name', 'Age', 'Occupation'],
    ['Alice', 25, 'Engineer'],
    ['Bob', 30, 'Doctor']
]

# Write to a CSV file
with open('people.csv', 'w', newline='') as file:
    writer = csv.writer(file)
    writer.writerows(rows)
```

[Reference: Writing to
CSV](https://docs.python.org/3/library/csv.html#csv.writer)

## Reading from a CSV File

To read data from a CSV file, use `csv.reader()`.

```python
import csv

# Read data from a CSV file
with open('people.csv', 'r') as file:
    reader = csv.reader(file)
    for row in reader:
        print(row)
```

[Reference: Reading from
CSV](https://docs.python.org/3/library/csv.html#csv.reader)

[Reference: Real Python - Reading and Writing CSV
Files](https://realpython.com/python-csv/)

## Context Managers

A context manager in Python is a mechanism that helps manage resources
efficiently, ensuring setup and cleanup tasks are handled automatically,
even in cases of exceptions. The `with` statement is commonly used with
context managers, ensuring that resources like files or network
connections are properly released after use.

### Example: Using a Context Manager with Files

```python
with open('example.txt', 'r') as file:
    content = file.read()
# The file is automatically closed after the 'with' block ends.
```

For more details, check out the official Python documentation on
[context
managers](https://docs.python.org/3/library/stdtypes.html#context-manager-types).

## Overriding the `str` Method

The `str` method in a Python class provides a human-readable string
representation of an object. This method is automatically invoked when
`print()` or `str()` is called on an instance of the class.

By default, the `str` method returns a generic and uninformative string
(like `<MyClass object at 0x...>`), but you can override it to make the
output more meaningful and informative.

### Example: Overriding `str`

```python
class Person:
    def __init__(self, name, age):
        self.name = name
        self.age = age

    def __str__(self):
        return f'Person(name={self.name}, age={self.age})'

# Usage
person = Person('Alice', 30)
print(person)  # Output: Person(name=Alice, age=30)
```

By overriding `str`, you control how objects of your class are presented
when printed. This is especially useful for debugging and logging. See
the Python documentation for more details on the [`str`
method](https://docs.python.org/3/reference/datamodel.html#object.__str__).



# Optional Extensions

### Custom Context Managers
[Optional for COMP150]

You can create a custom context manager by defining two special methods
in a class: `enter()` and `exit()`.

-   `enter()`: Executes when entering the `with` block and can return an
    object for use within the block.

-   `exit(exc_type, exc_value, traceback)`: Executes on exiting the
    `with` block, ensuring cleanup actions (such as closing a file or
    releasing a lock).

### Example: Custom Context Manager

```python
class MyContextManager:
    def __enter__(self):
        print("Entering context")
        return self  # The object returned here can be used in the 'with' block

    def __exit__(self, exc_type, exc_value, traceback):
        print("Exiting context")

# Usage
with MyContextManager() as cm:
    print("Inside context")
```

For a detailed guide, see the [Real Python tutorial on context
managers](https://realpython.com/python-with-statement/).

# Working with SQLite3 in Python

[Optional for COMP150]

SQLite is a lightweight database, and Python provides a built-in
`sqlite3` module to interact with it.

## Connecting to an SQLite Database

To connect to an SQLite database, use `sqlite3.connect()`.

```python
import sqlite3

# Connect to the database (or create it)
conn = sqlite3.connect('example.db')

# Create a cursor object
cursor = conn.cursor()
```

[Reference: Connecting to
Databases](https://docs.python.org/3/library/sqlite3.html#sqlite3.connect)

## Creating a Table

To create a table, use SQL `CREATE TABLE` statements.

```python
# SQL query to create a table
cursor.execute('''CREATE TABLE IF NOT EXISTS users
                  (id INTEGER PRIMARY KEY, 
                   name TEXT, 
                   age INTEGER)''')

# Commit changes
conn.commit()
```

[Reference: Creating Tables with
`sqlite3`](https://docs.python.org/3/library/sqlite3.html#sqlite3.Cursor.execute)

## Inserting Data into a Table

To insert data, use `INSERT INTO` statements.

```python
# Insert a row of data
cursor.execute("INSERT INTO users (name, age) VALUES (?, ?)", ('Alice', 25))

# Commit the transaction
conn.commit()
```

[Reference: Real Python - Using SQLite
Databases](https://realpython.com/python-sql-libraries/#sqlite)

## Querying Data from a Table

To retrieve data, use the SQL `SELECT` statement.

```python
# Query all rows from the users table
cursor.execute("SELECT * FROM users")

# Fetch all results
rows = cursor.fetchall()
for row in rows:
    print(row)
```

[Reference: Fetching Data from
SQLite](https://docs.python.org/3/library/sqlite3.html#sqlite3.Cursor.fetchall)

## Updating Data

To update data, use the SQL `UPDATE` statement.

```python
# Update the age of a user
cursor.execute("UPDATE users SET age = ? WHERE name = ?", (26, 'Alice'))
conn.commit()
```

## Deleting Data

To delete data, use the SQL `DELETE` statement.

```python
# Delete a user
cursor.execute("DELETE FROM users WHERE name = ?", ('Bob',))
conn.commit()
```

## Closing the Connection

After finishing with the database, close the connection.

```python
conn.close()
```

[Reference: Python `sqlite3`
Module](https://docs.python.org/3/library/sqlite3.html)

