# Introduction to Data Structures: Numpy, Lists, Arrays, and Tuples


NumPy is a powerful Python library for numerical computing, designed for efficient manipulation of large datasets. See also:
- [NumPy documentation](https://numpy.org/doc/stable/) It provides two primary data structures: **arrays** and **matrices**. This guide introduces these structures, assuming familiarity with basic Python concepts, but no prior knowledge of NumPy.

## NumPy Arrays

A **NumPy array** is a multi-dimensional, homogeneous collection of elements, with all items of the same data type. Arrays are the fundamental building block for many numerical computations in Python.

### Array Creation

To create an array:

See also:
- [NumPy Array Creation Guide](https://numpy.org/doc/stable/user/basics.creation.html)

```python
import numpy as np

# Creating a 1D array
a = np.array([1, 2, 3])

# Creating a 2D array
b = np.array([[1, 2], [3, 4]])
```

### Array Slicing and Aggregates

NumPy arrays support slicing, allowing selection of subsets of data.

See also:
- [NumPy Indexing and Slicing Guide](https://numpy.org/doc/stable/reference/arrays.indexing.html)

```python
# Slicing a 2D array
subset = b[:, 1]  # Select all rows from the second column
```

Arrays also provide efficient aggregate functions.

```python
# Aggregate functions
sum_all = np.sum(b)            # Sum of all elements
mean_value = np.mean(a)        # Mean of the array
dim_sum = np.sum(b, axis=0)    # Sum along columns
```

## NumPy Matrices

A **NumPy matrix** is a specialized 2D structure, a subclass of an array. It is mostly used for linear algebra operations.

See also:
- [NumPy Matrix Documentation](https://numpy.org/doc/stable/reference/generated/numpy.matrix.html)

```python
# Creating a matrix
m = np.matrix([[1, 2], [3, 4]])

# Matrix multiplication
result = m * m.T
```

Note: The use of matrices in NumPy is being deprecated in favor of multi-dimensional arrays. Arrays are more flexible and can perform equivalent operations.

## Performance Characteristics

**NumPy arrays** are implemented in C, providing significant performance improvements over Python lists due to:

- **Contiguous memory allocation**.
- **Vectorized operations**, which minimize loops and leverage optimized C functions.

Arrays are faster than matrices because they support efficient, generalized multi-dimensional operations and reduce overhead when performing element-wise operations.

### Optimizing Calculations with NumPy

NumPy's performance benefits can be maximized by using **vectorized operations** rather than Python loops.

See also:
- [NumPy Performance Tips](https://numpy.org/doc/stable/user/absolute_beginners.html#performance)

Consider calculating the element-wise sum of two arrays:

```python
# Using Python loops
size = 1000000
list1, list2 = range(size), range(size)
result = [list1[i] + list2[i] for i in range(size)]

# Using NumPy vectorization
a1 = np.arange(size)
a2 = np.arange(size)
result = a1 + a2  # Fast due to vectorized addition
```

The NumPy version is significantly faster as it leverages low-level optimizations. To maximize performance, plan calculations to avoid explicit loops, and use array slicing and aggregate functions whenever possible.

## Loading Data from a CSV File

NumPy can load numerical data directly from text files, such as CSVs, without using additional libraries like pandas.

Assume you have a CSV file `data.csv` with the following content:

```
Value,Label
1,A
2,B
3,C
4,D
```

To load only the numerical data:

```python
# Loading numerical data from a CSV file
import numpy as np

# Use genfromtxt to load the numerical data, skipping the header row and ignoring the text column
data = np.genfromtxt('data.csv', delimiter=',', skip_header=1, usecols=0)

print(data)  # Output: [1. 2. 3. 4.]
```

In this example, `np.genfromtxt()` is used to read the CSV file, skipping the header and selecting only the numerical column (`usecols=0`). This allows you to load numerical data efficiently without needing to use pandas.

## Comparison with Python Arrays, Lists, and Tuples

### Python Arrays

Python provides the `array` module to create arrays. These arrays are homogeneous like NumPy arrays but lack the advanced capabilities, performance, and flexibility that NumPy offers. Python arrays are useful for simple applications where a lightweight solution is sufficient. However, they do not support multi-dimensional data or efficient vectorized operations.

**Key Differences:**
- **Performance**: NumPy arrays are faster because they are implemented in C and optimized for numerical operations.
- **Functionality**: Python arrays have limited capabilities compared to NumPy arrays, which offer extensive mathematical functions, slicing, and broadcasting.

**Use Case**: If you only need basic one-dimensional storage without advanced operations, Python arrays are a simpler choice. NumPy arrays are more suitable for numerical computations.

### Python Tuples

### Python Lists

Python **lists** are general-purpose, ordered collections that can hold heterogeneous elements. Lists are mutable, allowing their contents to be changed, and they support a wide range of operations. However, they are not optimized for numerical computations like NumPy arrays.

**Key Differences:**
- **Performance**: Lists are slower compared to NumPy arrays for numerical operations because they are not implemented in a low-level language like C.
- **Data Type**: Lists can hold mixed data types, while NumPy arrays are homogeneous, making them more efficient for numerical data.
- **Functionality**: NumPy arrays support vectorized operations, slicing, and broadcasting, while lists require explicit iteration for element-wise operations.

**Use Case**: Use lists for small datasets with mixed types or when you need the flexibility of a mutable, general-purpose container. For numerical computations, NumPy arrays are a more efficient choice.

Tuples are immutable, ordered collections in Python that can store heterogeneous elements. Unlike NumPy arrays, they are not optimized for numerical operations and cannot be used directly for mathematical computations. Tuples are useful for storing related data of different types.

**Key Differences:**
- **Mutability**: Tuples are immutable, meaning their contents cannot be changed after creation, whereas NumPy arrays are mutable.
- **Data Type**: Tuples can hold mixed data types, whereas NumPy arrays are homogeneous.
- **Numerical Operations**: NumPy arrays are designed for mathematical operations, while tuples are general-purpose collections.

**Use Case**: Use tuples when you need a fixed collection of heterogeneous data that won’t change. For efficient mathematical computations, NumPy arrays are far superior.

## Optional Extension: Plotting with Seaborn

For visualizing statistical data, **Seaborn** is a powerful Python library that works well with NumPy. See also:
- [Seaborn Documentation](https://seaborn.pydata.org/). Below is an example of using Seaborn to create a plot from some statistical data.

First, install Seaborn if it is not already installed:

See also:
- [Matplotlib Documentation](https://matplotlib.org/stable/contents.html)

```sh
pip install seaborn
```

### Plotting Example

Assume we have some data loaded using NumPy, and we want to visualize it.

See also:
- [Scikit-learn Documentation (for advanced analysis)](https://scikit-learn.org/stable/)

```python
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt

# Example data
data = np.random.normal(size=100)

# Create a histogram plot
sns.histplot(data, kde=True)

# Show the plot
plt.show()
```

In this example:

- We use `np.random.normal()` to generate 100 random data points from a normal distribution.
- `sns.histplot()` is used to create a histogram of the data, with a Kernel Density Estimate (KDE) overlay.
- Finally, `plt.show()` displays the plot.

Seaborn provides a convenient and attractive way to visualize data, helping to understand the underlying statistical properties.




